#ifndef __ERROR_H__
#define __ERROR_H__

#include <string>
#include "FdsfeApi.h"

namespace FactSet {
    namespace Datafeed {
        struct ErrorImpl;
        class FDSFE_API Error {
        public:            
            /*! @brief Set of possible error codes.
             */
            enum ErrorCode {
                // Any changes made to this enum should be duplicated in ../../cpp_toolkit_c_sharp_wrapper/Error.cs
                // and ../../cpp_toolkit_c_wrapper/include/Error_C.h
                FE_NO_ERROR,    //!< The operation completed successfully and no error has occurred
                FE_E_UNKNOWN,   //!< Unexpected failure due to unknown reasons
                FE_E_NO_SERV,   //!< The requested service was unavailable
                FE_E_NOT_FOUND, //!< A required resource was not found
                FE_E_RENAME,    //!< A required resource was renamed
                FE_E_EXISTS,    //!< The output of this operation already exists
                FE_E_LIMIT,     //!< A resource limit was reached
                FE_E_PROTOCOL,  //!< An error occurred in the communication protocol
                FE_E_INVAL,     //!< The request was invalid
                FE_E_RESOURCE,  //!< A required resource was exhausted
                FE_E_NO_CONN,   //!< There was no connection to the server
                FE_E_VERSION,   //!< The operation was invalid for the current version
                FE_E_SHUTDOWN,  //!< Operation failed because the system is shutting down
                FE_E_ACCESS,    //!< The user lacked the necessary permissions
                FE_E_TIMEOUT,   //!< The operation timed out
                FE_E_UNEXPECTED_SUBSCRIBE,  //!< An unexpected subscribe response was received
                FE_E_UNEXPECTED_UNSUBSCRIBE //!< An unexpected unsubscribe response was received
            };

            /*! @brief Error indicating that no error has occurred.
             */
            const static Error NoError;


            /*! @brief Constructor
             * 
             * Constructs a new error with error code Error::FE_NO_ERROR and an empty string for the description.
             */
            Error();

            /*! @brief Constructor
            *
            * Constructs a new error with the given error code and description.
            *
            * @param code The Error::ErrorCode of the error
            * @param description The text description of the error
            */
            Error(ErrorCode code, const std::string &description);

            Error(const Error& other);

            Error& operator=(const Error& other);

            /*! @brief Destructor
            */
            ~Error();

            /*! @brief Gets the Error::ErrorCode of the Error.
             */
            ErrorCode code() const;

            /*! @brief Gets the text description of the Error.
             */
            const std::string& description() const;

            /*! @brief Equals operator override.
            *
            * Errors are considered equal if they have the same Error::ErrorCode and description.
            *
            * @param rhs The right hand side of the == operator.
            * @returns True if this Error is equal to the second Error.
            */
            bool operator==(const Error &rhs) const;

            /*! @brief Not equals operator override.
            *
            * @param rhs The right hand side of the != operator.
            * @returns True if this Error is not equal to the second Error.
            *
            * @see operator==()
            */
            bool operator!=(const Error &rhs) const;
        private:  
            ErrorImpl *m_impl;
        };
    } // namespace FactSet
} // namespace Datafeed
#endif
