#ifndef __EVENT_H__
#define __EVENT_H__

#include <string>
#include <vector>
#include "Error.h"
#include "FdsfeApi.h"

namespace FactSet {
    namespace Datafeed {
        struct EventImpl;
        class FDSFE_API Event {
        public:            
            /*! @brief Set of possible event types.
             *
             * Events are most commonly spawned to indicate the completion of the associated @ref asynchronous_methods "asynchronous method".
             * Events and their associated methods are listed below.
             */
          // Any changes made to this enum should be duplicated in ../../cpp_toolkit_c_sharp_wrapper/Event.cs
          // and ../../cpp_toolkit_c_wrapper/include/Event_C.h
          enum EventType {
              // Events are spawned after:
              STOP,                //!< FEConsumer::stop_async()
              AUTHENTICATE,        //!< FEConsumer::authenticate_async()
              CONNECTION_CHANGED,  //!< This Event can be spawned independently
                                   //!< of any asynchronous function call.
                                   //!< Whenever the FEConsumer was connected
                                   //!< and becomes disconnected or was
                                   //!< disconnected and becomes connected, this
                                   //!< Event will be spawned. Note that this is
                                   //!< true even if the synchronous interface
                                   //!< is used.
              CONNECTION_RECOVERY, //!< Similar to CONNECTION_CHANGED, this
                                   //!< event will be spawned after an attempt
                                   //!< to recover a lost connection to the
                                   //!< server to indicate the success of that
                                   //!< attempt
              CONNECT,             //!< FEConsumer::connect_async()
              DISCONNECT,          //!< FEConsumer::disconnect_async()
              LOGIN,               //!< FEConsumer::log_in_async()
              SUBSCRIBE,           //!< FEConsumer::subscribe_async()
              UNSUBSCRIBE,         //!< FEConsumer::unsubscribe_async()
              REQUEST_FILE,        //!< FEConsumer::request_file_async()
              SNAPSHOT_COMPLETE,   //!< FEConsumer::subscribe() or FEConsumer::subscribe_async() with snapshots
              RECOVER,             //!< FEConsumer::recover_async()
              RECOVERY_COMPLETE,   //!< FEConsumer::recover() or FEConsumer::recover_async()
              REQUEST_SNAPSHOT,    //!< FEConsumer::request_snapshot_async()
              COUNT                //!< Dummy value indicating number of EventTypes
          };

          /*! @brief Gets the string representation of the given
           * Event::EventType enum.
           */
          static std::string to_string(EventType type);

          Event();

          Event(const Event& other);

          Event& operator=(const Event& other);

          ~Event();

          /*! @brief Gets the Error associated with this event.
           *
           * @returns The Error that occured or Error::NoError if the operation
           * was a success.
           */
          const Error& error() const;

          /*! @brief Gets the EventType of this Event.
           */
          EventType type() const;

          /*! @brief Gets the unique id of this Event.
           */
          const std::string& id() const;

          /*! @brief Checks if a value with the given key exists in the Event.
           *
           * @returns True if the value exists, false otherwise.
           */
          bool has_value(const std::string& key) const;

          //Changes to this documentation should be copied over to the C# Wrapper Event.cs and C wrapper Event_C.h
          //Really, all documentation changes should, but this one seems important
          /*! @brief Gets the value for the given key.
           *
           * Some events will contain additional information depending on the
           * type of the event. See below for tables containing keys by event
           * type and their meanings.
           *
           * #AUTHENTICATE
           * | Key | Type | Description |
           * | --- | ---- | ------- |
           * | Host | std::string | Value of host argument from request. |
           * | User | std::string | Value of user argument from request. |
           * | Serial | std::string | Value of serial argument from request. |
           * | KeyId | std::string | Value of key_id argument from request. |
           * | Key | std::string | Value of key argument from request. |
           * | Counter | std::string | Value of counter argument from request. |
           * | Path | std::string | Value of path argument from request. |
           * | ForceInput | bool | Value of force_input argument from request. |
           *
           * #SUBSCRIBE
           * | Key | Type | Description |
           * | --- | ---- | ------- |
           * | Topic | std::string | Value of topic argument from request. |
           * | SnapshotMode | FEConsumer::SnapshotMode | Value of snapshot_mode argument from request. |
           * | DataMode | FEConsumer::DataMode | Value of data_mode argument from request. |
           *
           * #UNSUBSCRIBE
           * | Key | Type | Description |
           * | --- | ---- | ------- |
           * | Topic | std::string | Value of topic argument from request. |
           *
           * #CONNECTION_CHANGED
           * | Key | Type | Description |
           * | --- | ---- | ------- |
           * | Connected | bool | True if the FEConsumer is currently connected, false otherwise. |
           *
           * #CONNECTION_RECOVERY
           * | Key | Type | Description |
           * | --- | ---- | ------- |
           * | LastHeartbeat | std::chrono::system_clock::rep | Timestamp of the last heartbeat received in number of ticks since epoch. |
           *
           * #LOG_IN
           * | Key | Type | Description |
           * | --- | ---- | ------- |
           * | RequestFiles | bool | Value of request_files argument from request. |
           *
           * #REQUEST_FILE
           * | Key | Type | Description |
           * | --- | ---- | ------- |
           * |Filename | string | Value of filename argument from request. |
           *
           * #SNAPSHOT_COMPLETE
           * | Key | Type | Description |
           * | --- | ---- | ------- |
           * | Topic | std::string | Value of topic argument from request. |
           * | SnapshotCount | uint32_t | Number of snapshot messages received. |
           *
           * @param key Key to query.
           * @returns Value for key or empty string if there is no value.
           */
          const std::string& get_value(const std::string& key) const;

          /*! @brief Gets all key-value pairs in the event.
           *
          * @returns A string containing a list of all key value pairs.
          */
          const std::string& all_values() const;

        private:
            friend struct EventImpl;
            EventImpl *m_impl;
        };
    } // namespace FactSet
} // namespace Datafeed
#endif