#ifndef __FECONSUMER_H__
#define __FECONSUMER_H__

#include <string>
#include "Error.h"
#include "Event.h"
#include "FEConsumerOptions.h"
#include "FdsfeApi.h"
#include "LogMessage.h"
#include "MD/MD_mInt.h"

namespace FactSet {
namespace Datafeed {
struct FEConsumerImpl;

/*! @brief FEConsumer is a class which serves as the primary interface for
 * interacting with the broadcast server.
 *
 */
class FDSFE_API FEConsumer {
  public:
    //! @brief Operating system file descriptor type.
    #ifdef __linux__ 
    typedef int SocketType;
    #elif _WIN32
      typedef SOCKET SocketType;
    #endif

    //! @brief Market data message type.
    typedef MD::MD_Message MsgType;

    /*! @brief Callback signature for receiving market data messages.
     *
     *  @param topic The topic of msg.
     *  @param msg The message being received.
     *  @param closure The closure that was given when the callback was set or
     *  NULL if no closure was given.
     */
    typedef void (*MessageCallback)(const std::string& topic,
                                    MsgType& msg,
                                    void* closure);

    /*! @brief Callback signature for receiving events.
     *
     *  @param e The event being received.
     *  @param closure The closure that was given when the callback was set or
     * NULL if no closure was given.
     *
     *  @see event_cb()
     */
    typedef void (*EventCallback)(const Event& e, void* closure);

    /*! @brief Callback signature for receiving logging messages
     *
     *  For use with FEConsumer::set_logging_callback() in order to integrate with
     *  existing logging frameworks.
     *
     *  @param messages Array of logging messages returned
     *  @param count Length of messages array
     */
    typedef void(*LoggingCallback)(const LogMessage *messages, size_t count);

    //! @brief Value which indicates commands should never time out.
    //Changes to this value need to be duplicated in ../../cpp_toolkit_c_sharp_wrapper/FEConsumer.cs
    //and ../../cpp_toolkit_c_wrapper/FEConsumer_C.h
    const static int NO_TIMEOUT = 0;

    //! @brief Value which indicates no messages should ever be dropped due to
    //! reaching the high watermark.
    //Changes to this value need to be duplicated in ../../cpp_toolkit_c_sharp_wrapper/FEConsumer.cs
    //and ../../cpp_toolkit_c_wrapper/FEConsumer_C.h
    const static int HWM_UNLIMITED = 0;

    /*! @brief Set of possible snapshot modes when subscribing.
     */
    enum SnapshotMode {
        //Any changes made to this enum should be duplicated in ../../cpp_toolkit_c_sharp_wrapper/FEConsumer.cs
        //  and ../../cpp_toolkit_c_wrapper/include/FEConsumer_C.h
        NO_SNAPSHOT,      //!< No snapshots will be requested
        REQUIRE_SNAPSHOT  //!< Snapshots will be requested
    };

    /*! @brief Set of possible data modes when subscribing.
     */
    enum DataMode {
        //Any changes made to this enum should be duplicated in ../../cpp_toolkit_c_sharp_wrapper/FEConsumer.cs
        //  and ../../cpp_toolkit_c_wrapper/include/FEConsumer_C.h
        LIVE,    //!< Live realtime market data will be requested
        DELAYED, //!< Delayed market data will be requested
        CANNED   //!< Prerecorded canned data will be requested
    };

    /*! @brief Set of possible data sources when registering callbacks.
    */
    enum DataSource {
        //Any changes made to this enum should be duplicated in ../../cpp_toolkit_c_sharp_wrapper/FEConsumer.cs
        //  and ../../cpp_toolkit_c_wrapper/include/FEConsumer_C.h
        MARKET,   //!< Market data from subscribe
        RECOVERY, //!< Recovery data from recover
        OUT_OF_BAND_SNAPSHOT, //!< Snapshot data from request_out_of_band_snapshot
        COUNT     //!< Dummy value indicating number of DataSources
    };

    /*! @brief Opens a file with the given filename to log messages from the
     * toolkit.
     *
     * @param filename The name of the file to store the logs.
     * @param append If true, appends to previous log file instead of creating a
     * fresh one.
     * @return Error if something went wrong or Error::NoError otherwise.
     */
    static Error log_open(const std::string& filename, bool append);

    /*! @brief Closes the log file previously opened with log_open().
     */
    static void log_close();

    /*! @brief Turn on logging additional messages.
     */
    static void enable_verbose_logging();

    /*! @brief Turn off logging additional messages.
     */
    static void disable_verbose_logging();

    /*! @brief Sends logging messages to function instead of file/console.
    *
    * @param callback LoggingCallback to call on log message.
    */
    static void set_logging_callback(LoggingCallback callback);

    /*! @brief Connected state of the FEConsumer.
     *
     * The FEConsumer is only considered connected if it is connected to the
     * data server on all channels. It is possible for this function to return
     * false while data is still being received if only one channel is
     * disconnected.
     *
     *  @return True if the FEConsumer is connected.
     *
     *  @see connect() @see connect_async()
     */
    bool connected() const;

    /*! @brief Authenticated state of the FEConsumer.
     *
     *  @return True if the FEConsumer is authenticated.
     *
     *  @see authenticate() @see authenticate_async()
     */
    bool authenticated() const;

    /*! @brief Login state of the FEConsumer.
     *
     *  @return True if the FEConsumer is logged in.
     *
     *  @see log_in() @see log_in_async()
     */
    bool logged_in() const;

    /*! @brief Checks for a subscription to a topic.
    *
    *  @param topic The topic to check for subscribed status.
    *  @return True if the FEConsumer is subscribed to topic.
    *
    *  @see subscribe() @see subscribe_async()
    */
    bool is_subscribed(const std::string& topic) const;

    /*! @brief Gets the seconds to spend on a single command before returning a
     * timeout error.
     *
     * @see timeout_sec(unsigned int seconds);
     */
    unsigned int timeout_sec() const;

    /*! @brief Gets the seconds to wait to receive all snapshot messages before
    * returning a timeout error.
    *
    * @see snapshot_timeout_sec(unsigned int seconds);
    */
    unsigned int snapshot_timeout_sec() const;

    /*! @brief Gets the seconds to wait to receive a response after sending a heartbeat
    * before considering the connection as timed out.
    *
    * @see heartbeat_timeout_sec(unsigned int seconds);
    */
    unsigned int heartbeat_timeout_sec() const;

    /*! @brief Gets the seconds to wait between sending heartbeat queries to the server
    *
    * @see heartbeat_interval_sec(unsigned int seconds);
    */
    unsigned int heartbeat_interval_sec() const;

    /*! @brief Gets the maximum number of messages to hold for each topic while
     * waiting for snapshots.
     *
     * @see max_snapshot_queue_size(unsigned int size)
     */
    unsigned int max_snapshot_queue_size() const;

    /*! @brief Sets the EventCallback() used for receiving events.
     *
     * If the closure is set, then it will be returned in the callback,
     * otherwise NULL will be returned.
     *
     * @param callback The EventCallback() to call when an event is received.
     */
    void event_cb(EventCallback callback);

    /*! @copybrief event_cb(EventCallback callback)
     *
     * @copydetails event_cb(EventCallback callback)
     * @param closure The closure to return in the EventCallback()
     */
    void event_cb(EventCallback callback, void* closure);

    /*! @brief Sets the seconds to spend on a single command before returning a
     * timeout error.
     *
     * Use #NO_TIMEOUT to indicate infinite timeout.
     *
     * @param seconds The number of seconds to wait.
     */
    void timeout_sec(unsigned int seconds);

    /*! @brief Sets the seconds to wait to receive all snapshot messages before
    * returning a timeout error.
    *
    * @param seconds The number of seconds to wait.
    */
    void snapshot_timeout_sec(unsigned int seconds);

    /*! @brief Sets the seconds to wait to receive a response after sending a heartbeat
    * before considering the connection as timed out.
    *
    * @param seconds The number of seconds to wait.
    */
    void heartbeat_timeout_sec(unsigned int seconds);

    /*! @brief Sets the seconds to wait between sending heartbeats.
    *
    * @param seconds The number of seconds to wait.
    */
    void heartbeat_interval_sec(unsigned int seconds);

    /*! @brief Sets the maximum number of messages to hold for each topic while
     * waiting for snapshots.
     *
     * When the maximum queue size is reached, older messages will be dropped in
     * favor of new ones.
     *
     * @param size The new max number of messages for the snapshot queue
     *
     * @see subscribe() @see subscribe_async()
     */
    void max_snapshot_queue_size(unsigned int size);

    /*! @brief Enables tracking message latency throughout the consumer.
    *
    * If enabled, the consumer will measure the latency for each message
    * at various internal checkpoints. Two latencies will be tracked: one
    * based on the MSG_GENESIS of the message (the time FactSet received
    * the message from the exchange, referred to as the "real latency")
    * and another based on the time the consumer received the message from
    * the broadcast server (referred to as the "toolkit latency"). The highest
    * latency observed during the bucket interval will be taken as the
    * latency for that interval.The latencies at each checkpoint will be
    * printed to the log at the end of the interval, in the following format:
    * CheckpointName [Real Latency (Toolkit Latency)]
    *
    * \b Checkpoints
    * | Name | Description |
    * | --- | ------- |
    * | Subscriber | The message was read off the socket from the broadcast server. |
    * | Worker | The message was received by a worker thread. |
    * | Callback Queue | The message was inserted into the matching callback queue. |
    * | Dispatch | The callback for the message was called during dispatching. |
    *
    * In addition to message latency, the consumer will also
    * track the time it takes to call all callbacks during a call to dispatch()
    * or dispatch_fd(). The longest of these durations within a bucket interval
    * will be printed as the callback time.
    *
    * All printed durations will be given in milliseconds.
    *
    * @param bucket_interval_sec The bucket interval in seconds.
    * @see disable_latency_tracking()
    */
    void enable_latency_tracking(unsigned int bucket_interval_sec);

    /*! @brief Disables tracking message latency throughout the consumer.
    *
    * @see enable_latency_tracking()
    */
    void disable_latency_tracking();

    /*! @brief Registers a callback for messages with topics starting with the
     * given prefix.
     *
     * For each registered callback, an internal notify queue will be created.
     * The file descriptor returned in the parameter fd_out will become readable
     * when a notify is triggered on this queue. A notify will be triggered
     * after a certain number of messages have been enqueued or after a certain
     * number of milliseconds have passed. After a notify is triggered, the file
     * descriptor should not be read from directly. Rather, dispatch_fd() should
     * be called to call the given FEConsumer::MessageCallback for each queued
     * message.
     *
     * @see dispatch()
     * @see set_notify_queue_message_limit()
     * @see set_notify_queue_time_limit_ms()
     *
     * @param topic_prefix The topic prefix associated with this callback.
     * @param callback The FEConsumer::MessageCallback to call when dispatched.
     * @param closure The closure to return in the MessageCallback.
     * @param fd_out On success, fd_out will be set to the file descriptor for
     * this callback.
     * @return Error if an error occurs, otherwise Error::NoError.
     */
    Error register_callback(const std::string&          topic_prefix,
                            FEConsumer::MessageCallback callback,
                            void*                       closure,
                            SocketType&                 fd_out);

    Error register_callback(const std::string& topic_prefix,
                            FEConsumer::MessageCallback callback,
                            void* closure,
                            DataSource source,
                            SocketType& fd_out);

    /*! @brief Unregisters a callback previously registered with
     * register_callback().
     *
     * @param topic_prefix The topic prefix of the callback to unregister.
     * @return Error if an error occurs, otherwise Error::NoError.
     */
    Error unregister_callback(const std::string& topic_prefix);
    Error unregister_callback(const std::string& topic_prefix,
                              DataSource source);

    /*! @copybrief unregister_callback(const std::string& topic_prefix)
     *
     * @param fd The file descriptor of the callback to unregister.
     * @return Error if an error occurs, otherwise Error::NoError.
     */
    Error unregister_callback(SocketType fd);

    /*! @brief Returns the file descriptor for a given topic prefix.
     *
     * The file descriptor returned will be the same value that was returned
     * when the callback was initially registered.
     * @see register_callback()
     *
     * @param topic_prefix The topic prefix for which to retrieve the file
     * descriptor.
     * @param fd_out On success, fd_out will be set to the file descriptor for
     * the callback.
     * @return Error if an error occurs, otherwise Error::NoError.
     */
    Error get_notify_socket(const std::string& topic_prefix,
                            SocketType& fd_out);

    Error get_notify_socket(const std::string& topic_prefix,
                            FEConsumer::DataSource source,
                            SocketType& fd_out);

    /*! @brief Convenience function for dispatching on all registered callbacks.
     *
     * The file descriptors for each registered callback will be polled and
     * dispatch_fd() will be called for each callback that has messages in its
     * queue. If no messages are immediately available when this function is
     * called, it will wait for up to timeout_ms milliseconds for messages to
     * become available before returning. If a timeout occurs, this is not
     * considered an error and Error::NoError will be returned.
     *
     * @param timeout_ms Number of milliseconds to wait before timing out.
     * @return Error if an error occurs, otherwise Error::NoError.
     */
    Error dispatch(long timeout_ms);

    /*! @brief Dispatches messages for the callback with the given file
     * descriptor.
     *
     * For each message in the notify queue for the callback with the given file
     * descriptor, the associated callback that was registered by
     * register_callback() will be called. This function should only be called
     * on a readable file descriptor.
     *
     * @param fd File descriptor to dispatch.
     * @return Error if an error occurs, otherwise Error::NoError.
     */
    Error dispatch_fd(SocketType fd);

    /*! @brief Sets the notify queue message limit.
     *
     * A notify queue that has not been notified will be notified when it
     * contains at least as many messages as the limit.
     *
     * @see register_callback()
     * @see set_notify_queue_time_limit_ms()
     *
     * @param limit The message limit count.
     */
    void set_notify_queue_message_limit(size_t limit);

    /*! @brief Gets the notify queue message limit.
     *
     * @see set_notify_queue_message_limit()
     *
     * @return The message limit count.
     */
    size_t get_notify_queue_message_limit();

    /*! @brief Sets the notify queue time limit in milliseconds.
     *
     * A notify queue that has not been notified will be notified when it
     * contains at least 1 message and the time limit has passed.
     *
     * @see register_callback()
     * @see set_notify_queue_message_limit()
     *
     * @param limit The new time limit in milliseconds
     */
    void set_notify_queue_time_limit_ms(int limit);

    /*! @brief Gets the notify queue time limit.
     *
     * @see set_notify_queue_time_limit_ms()
     *
     * @return The time limit in milliseconds.
     */
    int get_notify_queue_time_limit_ms();

    /*! @brief Constructor.
     *
     * Market data messages will be received, processed, and delivered by a pool
     * of subscriber and worker threads.
     *
     * Each subscriber thread will manage its own connection to the broadcast
     * server. Whenever a subscription to a new topic is made, a subscriber is
     * selected to handle that subscription on a round robin basis. Messages for
     * a topic will always be received by the same subscriber thread to preserve
     * proper ordering. Whenever a subscribe receives a message, that message is
     * forwarded to a worker thread. The worker chosen to process the message
     * will be consistent for a single topic.
     *
     * Worker threads process the messages and then forward them to a notify
     * queue registered with register_callback(). These messages can then be
     * dispatched by dispatch() or dispatch_fd().
     *
     * If the application is unable to process the messages fast enough, they
     * will begin to queue up until the system runs out of memory. If this is
     * undesirable, the high watermark can be set to drop messages once a
     * certain amount have been queued.
     *
     * @param sub_count The number of subscriber threads to create.
     * @param worker_count The number of worker threads to create.
     * @param high_watermark The high watermark of the subscribers and workers.
     * @param enable_compression True if market data should be compressed on the wire.
     *
     */
    FEConsumer(int sub_count, int worker_count, int high_watermark, bool enable_compression);

    /*! @copybrief FEConsumer(int subscriberThreads, int workerThreads, int
     * highWatermark)
     *
     * @overload
     *
     * High watermark will be set to #HWM_UNLIMITED.
     * Enable compression will be set to true.
     *
     */
    FEConsumer(int sub_count, int worker_count, int high_watermark);

    /*! @copybrief FEConsumer(int subscriberThreads, int workerThreads, int
     * highWatermark)
     *
     * @overload
     *
     * High watermark will be set to #HWM_UNLIMITED.
     *
     */
    FEConsumer(int sub_count, int worker_count);

    FEConsumer(const FEConsumerOptions& options);

    ~FEConsumer();

    /*! @name Synchronous methods
     * A synchronous method will block until it completes.
     * All synchronous methods will return an Error if something went wrong or
     * Error::NoError otherwise. If the operation does not complete within
     * timeout_sec() seconds, a timeout Error will be returned. Synchronous
     * methods cannot be called from FEConsumer callbacks.
     */
    //!@{

    /*! @brief Does nothing.
     *
     * @deprecated FEConsumer will be cleaned up by the destructor.
     */
    Error stop();

    /*! @brief Authenticates with the given credentials.
     *
     * A counter file will be created to store the key and counter used for
     * authentication at the given path. This file will be named \<key_id>.data.
     * If a counter file already exists, the key and counter it contains will be
     * used instead of the key and counter parameters. Setting the force_input
     * flag to true will overwrite the counter file values with the given key
     * and counter.
     *
     * @param host The address of the authentication server in the format
     * host:port
     * @param user The name of the user
     * @param serial The serial associated with user
     * @param key_id The key id of the key being used for authentication
     * @param key The key associated with key_id for user
     * @param counter The current counter for key
     * @param path The location to store the counter file
     * @param force_input If true, key and counter parameters will be used
     * instead of the values in an existing counter file
     */
    Error authenticate(const std::string& host,
                       const std::string& user,
                       const std::string& serial,
                       const std::string& key_id,
                       const std::string& key,
                       const std::string& counter,
                       const std::string& path,
                       bool               force_input);

    /*! @copybrief authenticate()
     *
     * @overload
     *
     * @param host The address of the authentication server in the format
     * host:port
     * @param user The name and serial of the user in the format user-serial
     * @param key_id The key id of the key being used for authentication
     * @param key The key associated with key_id for user
     * @param counter The current counter for key
     * @param path The location to store the counter file
     * @param force_input If true, key and counter parameters will be used
     * instead of the values in an existing counter file
     */
    Error authenticate(const std::string& host,
                       const std::string& user,
                       const std::string& key_id,
                       const std::string& key,
                       const std::string& counter,
                       const std::string& path,
                       bool               force_input);

    /*! @brief Connects to the broadcast server.
     *
     * The FEConsumer must be authenticated before connecting.
     * @see authenticate() @see authenticate_async()
     */
    Error connect();

    /*! @brief Disconnects from the broadcast server.
     *
     * Closes the connection to the broadcast server by logging out if
     * logged_in() and disconnecting. To reconnect after disconnecting it is
     * necessary to begin again at authentication.
     *
     * The FEConsumer must be authenticated before disconnecting.
     * @see authenticate() @see authenticate_async()
     * @see connect() @see connect_async()
     */
    Error disconnect();

    /*! @brief Logs in to the broadcast server.
     *
     * The FEConsumer must be connected before logging in.
     * @see connect() @see connect_async()
     */
    Error log_in(bool request_files = true);

    /*! @brief Subscribes to a topic.
     *
     * If snapshot messages were requested, all live data will be queued until
     * all snapshots are received for the requested topic. If
     * max_snapshot_queue_size() live data messages are queued before all
     * snapshots are received, the oldest of the queued messages will be dropped
     * in favor of newer messages. Once all snapshots are received, queued
     * messages which are newer than the snapshots will be delivered via the
     * registered callbacks as usual. Messages older than the snapshots will be
     * dropped.
     *
     * The format for the topic is productCode|issueType|isoCode|symbol.
     * It is possible to subscribe to a prefix of the format, for example a
     * subscription to a productCode|issueType|isoCode combination will
     * subscribe to all symbols that match the combination. It is not possible
     * to subscribe to two topics which overlap.
     *
     * @param topic The topic being subscribed.
     * @param snapshot_mode Determines whether or not snapshots will be
     * requested.
     * @param data_mode Determines the type of data to request.
     *
     * @see unsubscribe()
     * @see unsubscribe_async()
     * @see register_callback()
     *
     */
    Error subscribe(const std::string& topic,
                    SnapshotMode       snapshot_mode,
                    DataMode           data_mode);

    /*! @brief Unsubscribes from a topic.
     *
     * @param topic The topic being unsubscribed.
     *
     * @see subscribe() @see subscribe_async()
     *
     */
    Error unsubscribe(const std::string& topic);


    /*! @brief Requests a configuration file with a given name.
    *
    * @param filename The name of the file being requested.
    * @param contents_out The contents of the requested file or empty string if there was an error.
    *
    */
    Error request_file(const std::string& filename, std::string& contents_out);

    Error recover(const std::string& topic,
                  time_t begin_time,
                  time_t end_time,
                  FEConsumer::DataMode data_mode);

    Error request_snapshot(const std::string& topic,
                           FEConsumer::DataMode data_mode);

    //!@} // End doxygen Synchronous methods group

    /*! @name Asynchronous methods
     * @anchor asynchronous_methods
     * An asynchronous method will return immediately and will send an Event via
     * the event_cb() when it completes. If the operation does not complete
     * within timeout_sec() seconds, an Event containing a timeout Error will be
     * generated. Asynchronous methods can be called from FEConsumer callbacks.
     */
    //!@{

    /*! @copydoc stop()
     */
    void stop_async();

    /*! @copydoc authenticate(const std::string &host, const std::string &user,
     * const std::string &serial, const std::string &key_id, const std::string
     * &key, const std::string &counter, const std::string &path, bool
     * force_input)
     */
    void authenticate_async(const std::string& host,
                            const std::string& user,
                            const std::string& serial,
                            const std::string& key_id,
                            const std::string& key,
                            const std::string& counter,
                            const std::string& path,
                            bool               force_input);

    /*! @copydoc authenticate(const std::string &host, const std::string &user,
     * const std::string &key_id, const std::string &key, const std::string
     * &counter, const std::string &path, bool force_input)
     */
    void authenticate_async(const std::string& host,
                            const std::string& user,
                            const std::string& key_id,
                            const std::string& key,
                            const std::string& counter,
                            const std::string& path,
                            bool               force_input);

    /*! @copydoc connect()
     */
    void connect_async();

    /*! @copydoc disconnect()
     */
    void disconnect_async();

    /*! @copydoc log_in()
     */
    void log_in_async(bool request_files = true);

    /*! @copydoc subscribe(const std::string&,SnapshotMode,DataMode,DelayMode)
     */
    void subscribe_async(const std::string& topic,
                         SnapshotMode       snapshot_mode,
                         DataMode           data_mode);

    /*! @copydoc unsubscribe()
     */
    void unsubscribe_async(const std::string& topic);

    /*! @copydoc request_file()
     */
    void request_file_async(const std::string& filename);

    void recover_async(const std::string& topic,
                       time_t begin_time,
                       time_t end_time,
                       FEConsumer::DataMode data_mode);

    void request_snapshot_async(const std::string& topic,
                                FEConsumer::DataMode data_mode);

    //!@} // End doxygen Asynchronous methods group
#ifdef FDS_C_WRAPPER
    void create_hwm_event(Event::EventType type, const std::string& topic);
#endif

  private:
    FDSFE_LOCAL FEConsumer(const FEConsumer& not_copyable);
    FDSFE_LOCAL FEConsumer& operator=(const FEConsumer& not_copyable);
    friend struct FEConsumerImpl;
    FEConsumerImpl* m_impl;
};
}  // namespace Datafeed
}  // namespace FactSet
#endif
