#ifndef __MD_LIB_FLOATNORMALIZER_H__
#define __MD_LIB_FLOATNORMALIZER_H__
#pragma once

#include <set>
#include <climits>

#include <MD/private/prefix.h>
#include <MD/MD_FieldDefs.h>

#include <MD/MD_FieldMap.h>
#include <MD/MD_Fracker.h>

namespace MD
{
   // --------------------------------------------------------------------------------------------
   // used to convert Market Data messages between floats/double to whole/num/denom or vice-versa.
   // Also provides such operations on field level.
   //
   // This class can be considered safe to use in a multi-threaded environment, provided that 
   // initialization is done before any threads attempt to use it.  additionally, reinitialzing
   // requires that any threads accessing the singleton be paused.  this constraint allows us 
   // to avoid using mutexes in this code.  if for some reason you require the ability to initialize and/or 
   // reinitialize the singleton, you should write your own locking wrapper around this.
   // --------------------------------------------------------------------------------------------
   class MD_FloatNormalizer
   {
   private:
      typedef MD::FIELDS::FieldID FieldID;

   public:
      class LoadOptions
      {
         public:
            LoadOptions(void);

            LoadOptions &set_panic_on_error(bool enable);
            LoadOptions &set_use_file_resolver(bool enable);

            bool get_panic_on_error(void) const;
            bool get_use_file_resolver(void) const;

         private:
            bool m_panic_on_error;
            bool m_use_file_resolver;
      };

      // Set this prior to calling instance() or reinitialize() 
      // if you want to change the default filename of the mapping config file
      // The default filename is "FloatFieldMapping.txt" on Windows
      // and "floatnorm/FloatFieldMapping.txt" on Unix
      // returns the original file name
      // The loading options are overwritten with the options argument when it's
      // not NULL
      MD_LIB_API static std::string set_mapping_config_file( const std::string &filename, const LoadOptions *options = NULL );

      // THIS IS NOT THREAD SAFE. Please call instance() from your main thread before your spawn others.
      MD_LIB_API static const MD_FloatNormalizer & instance();

      // THIS IS NOT THREAD SAFE. Threads accessing an instance of this should be paused prior to use. 
      // The workstation needs this feature to reload the mapping file if it gets a new one 
      // from the server file check on handshake
      static const MD_FloatNormalizer & reinitialize();

      // returns true of the config file was loaded successfully, with at least one
      // entry added to the map.  does not account for invalid entries or missing data,
      // as long as there is at least one good entry in the file, we consider this
      // a valid load.
      bool is_valid() const { return m_is_valid; }

      // Returns the double value of the fid if already present in the msg.
      // Otherwise, it tries to get the w/n/d or float and return double.
      template<class MessageType>
      bool get_double(const MessageType &msg, const FieldID fid, double &out) const;

      // Returns the float value of the fid if already present in the msg
      // Otherwise, it tries to get the w/n/d or double and return float.
      template<class MessageType>
      bool get_float(const MessageType &msg, const FieldID fid, float &out) const;

      // Returns the corresponding double fid for the supplied fid
      MD_LIB_API bool get_double_fid(const FieldID fid, FieldID &double_fid) const;

      // Returns the corresponding float fid for the supplied fid
      MD_LIB_API bool get_float_fid(const FieldID fid, FieldID &float_fid) const;

      // Normalizes the message in place
      template<class MessageType>
      void normalize_message(const std::set<FieldID> &normalized_fids, MessageType &msg) const;

      // Normalizes a copy of the message
      template<class MessageType>
      void normalize_message(const std::set<FieldID> &normalized_fids, const MessageType &src, const MessageType &dest) const;

      // normalize all w/n/d in the message
      // to Double if it is a valid fid
      // else to a float fid if it is valid
      // Should not be used by ticker plants, for use by app only when they get rid of OPTIs
      template<class MessageType>
      void normalize_message(MessageType &msg) const;

      // Converts all double/float fields to w/n/d if fid is available.
      template<class MessageType>
      void denormalize_message(MessageType &msg) const;

      // Prints the full internal mapping
      MD_LIB_API void print(std::ostream &os, bool print_collapsed = false) const;

      // print a collapsed view of internal mapping
      MD_LIB_API void print_collapsed(std::ostream &os) const;

   private:
      MD_FloatNormalizer();
      MD_FloatNormalizer(const MD_FloatNormalizer & not_copyable);
      MD_FloatNormalizer & operator=(const MD_FloatNormalizer & not_copyable);

      ~MD_FloatNormalizer();

      void initialize();

      // This represents all fields which represent the same information or a part of it.
      struct FieldGroup
      {
         FieldGroup(const std::vector<FieldID> &info);

         void print(std::ostream &o) const;
         void print_collapsed(std::ostream &o) const;

         static const FieldID UNDEFINED_FID = 0xFFFF;
         static bool is_valid_fid(FieldID id) { return id != UNDEFINED_FID; }

         FieldID whole_fid;
         FieldID num_fid;
         FieldID denom_fid;
         FieldID float_fid;
#ifdef MD_LIB_WIN
         FieldID opti_fid;
#endif
         FieldID double_fid;
         bool    is_valid;
      };

      typedef std::vector<FieldGroup *> FidToFieldGroup_t;

      const FieldGroup * get_field_group(const FieldID fid) const;
      const FieldGroup * get_field_group_if_double(const FieldID fid) const;
      const FieldGroup * get_field_group_if_whole(const FieldID fid) const;

      bool m_is_valid;
      FidToFieldGroup_t m_fid_to_field_group;
      FidToFieldGroup_t m_double_fid_to_field_group;
      FidToFieldGroup_t m_whole_fid_to_field_group;
      const MD::MD_FieldMap& m_field_map;

      static MD_FloatNormalizer& instance_();

      void clearFieldGroups();
   };
} // namespace MD


// IMPLEMENTATION DETAILS

namespace MD
{
   inline MD_FloatNormalizer::LoadOptions::LoadOptions(void) : m_panic_on_error(true), m_use_file_resolver(true)
   {
   }

   inline MD_FloatNormalizer::LoadOptions &MD_FloatNormalizer::LoadOptions::set_panic_on_error(const bool enable)
   {
      m_panic_on_error = enable;
      return *this;
   }

   inline MD_FloatNormalizer::LoadOptions &MD_FloatNormalizer::LoadOptions::set_use_file_resolver(const bool enable)
   {
      m_use_file_resolver = enable;
      return *this;
   }

   inline bool MD_FloatNormalizer::LoadOptions::get_panic_on_error(void) const
   {
      return m_panic_on_error;
   }

   inline bool MD_FloatNormalizer::LoadOptions::get_use_file_resolver(void) const
   {
      return m_use_file_resolver;
   }
}

inline const MD::MD_FloatNormalizer::FieldGroup * MD::MD_FloatNormalizer::get_field_group(const FieldID fid) const
{
   return (fid < FieldGroup::UNDEFINED_FID ? m_fid_to_field_group[fid] : NULL);
}

inline const MD::MD_FloatNormalizer::FieldGroup * MD::MD_FloatNormalizer::get_field_group_if_double(const FieldID fid) const
{
   return (fid < FieldGroup::UNDEFINED_FID ? m_double_fid_to_field_group[fid] : NULL);
}

inline const MD::MD_FloatNormalizer::FieldGroup * MD::MD_FloatNormalizer::get_field_group_if_whole(const FieldID fid) const
{
   return (fid < FieldGroup::UNDEFINED_FID ? m_whole_fid_to_field_group[fid] : NULL);
}

// Can't instantiate these functions in cpp because they will be used by MarqueeServer and App
// for MDS::Message which does not derive from MD_Message but has the same interface.
template<class MessageType>
bool MD::MD_FloatNormalizer::get_double(const MessageType &msg, const FieldID fid, double &out) const
{
   // to avoid log errors due to mismatched types in MD_Message::getDouble
   const auto type = m_field_map.idToType(fid);
   if ((type == MD::DATATYPES::DOUBLE) && msg.getDouble(fid, out)) {
     return true;
   }

   const FieldGroup * group = get_field_group(fid);
   if (!group)
      return false;

   if (FieldGroup::is_valid_fid(group->double_fid) && msg.getDouble(group->double_fid, out))
      return true;

   if (FieldGroup::is_valid_fid(group->whole_fid) &&
      FieldGroup::is_valid_fid(group->num_fid) &&
      FieldGroup::is_valid_fid(group->denom_fid))
   {
      int denom_val = 0;
      if (msg.getInt(group->denom_fid, denom_val) && denom_val == 0)
      {
         // Division by zero, set output to zero
         out = 0;
         return true;
      }

      if (msg.getWND(group->whole_fid, group->num_fid, group->denom_fid, out))
         return true;
   }

   float float_val;
   if (FieldGroup::is_valid_fid(group->float_fid) && msg.getFloat(group->float_fid, float_val))
   {
      out = static_cast<double>(float_val);
      return true;
   }

   return false;
}

template<class MessageType>
bool MD::MD_FloatNormalizer::get_float(const MessageType &msg, const FieldID fid, float &out) const
{
   // to avoid log errors due to mismatched types in MD_Message::getFloat
   const auto type = m_field_map.idToType(fid);
   if ((type == MD::DATATYPES::FLOAT) && msg.getFloat(fid, out)) {
      return true;
   }

   const FieldGroup * group = get_field_group(fid);
   if (!group)
      return false;

   if (FieldGroup::is_valid_fid(group->float_fid) && msg.getFloat(group->float_fid, out))
      return true;

   double double_val;
   if (FieldGroup::is_valid_fid(group->whole_fid) &&
      FieldGroup::is_valid_fid(group->num_fid) &&
      FieldGroup::is_valid_fid(group->denom_fid))
   {
      int denom_val = 0;
      if (msg.getInt(group->denom_fid, denom_val) && denom_val == 0)
      {
         // Division by zero, set output to zero
         out = 0;
         return true;
      }

      if (msg.getWND(group->whole_fid, group->num_fid, group->denom_fid, double_val))
      {
         out = static_cast<float>(double_val);
         return true;
      }
   }

   if (FieldGroup::is_valid_fid(group->double_fid) && msg.getDouble(group->double_fid, double_val))
   {
      out = static_cast<float>(double_val);
      return true;
   }

   return false;
}

template<class MessageType>
void MD::MD_FloatNormalizer::normalize_message(const std::set<FieldID> &normalized_fids, MessageType &msg) const
{
   std::set<FieldID> fids_to_erase;

   for (std::set<FieldID>::const_iterator it = normalized_fids.begin(), it_e = normalized_fids.end(); it != it_e ; ++it)
   {
      const FieldID &fid(*it);
      bool normalized = false;

      // If the field is already present in the message or if is marked for erase, skip
      if (msg.exists(fid) || fids_to_erase.find(fid) != fids_to_erase.end())
         continue;

      const FieldGroup * group = get_field_group(fid);
      if (!group)
         continue;

      switch(m_field_map.idToType(fid))
      {
      case MD::DATATYPES::DOUBLE:
         {
            double val = 0;
            if (get_double(msg, fid, val))
            {
               if (FieldGroup::is_valid_fid(group->float_fid))
                  fids_to_erase.insert(group->float_fid);

               msg.addDouble(fid, val);
               normalized = true;
            }
            break;
         }
      case MD::DATATYPES::FLOAT:
         {
            float val = 0;
            if (get_float(msg, fid, val))
            {
               if (FieldGroup::is_valid_fid(group->double_fid))
                  fids_to_erase.insert(group->double_fid);

               msg.addFloat(fid, val);
               normalized = true;
            }
            break;
         }
      default:
         break;
      }

      if (normalized)
      {
         if (FieldGroup::is_valid_fid(group->whole_fid))
            fids_to_erase.insert(group->whole_fid);

         if (FieldGroup::is_valid_fid(group->num_fid))
            fids_to_erase.insert(group->num_fid);

         if (FieldGroup::is_valid_fid(group->denom_fid))
            fids_to_erase.insert(group->denom_fid);
#ifdef MD_LIB_WIN
         if (FieldGroup::is_valid_fid(group->opti_fid))
            fids_to_erase.insert(group->opti_fid);
#endif
      }
   }

   msg.erase(fids_to_erase);
}

template<class MessageType>
void MD::MD_FloatNormalizer::normalize_message(const std::set<FieldID> &normalized_fids,
                                           const MessageType &src, const MessageType &dest) const
{
   dest.copyMsg(src);
   normalize_message(normalized_fids, dest);
}

template<class MessageType>
void MD::MD_FloatNormalizer::normalize_message(MessageType &msg) const
{
   std::vector<std::pair<FieldID, double> > dbl_fid_vals_to_add;
   std::vector<std::pair<FieldID, float > > float_fid_vals_to_add;

   std::set<FieldID> fids_to_erase;
   std::vector<FieldID> all_fids;

   msg.get_field_ids(all_fids);

   int whole, num, denom;

   std::vector<FieldID>::const_iterator it = all_fids.begin(), it_e = all_fids.end();
   for (; it != it_e; ++it)
   {
      const FieldID &fid(*it);

      const FieldGroup *group = get_field_group_if_whole(fid);
      if (!group) continue;

      if (!FieldGroup::is_valid_fid(group->num_fid) || !FieldGroup::is_valid_fid(group->denom_fid))
         continue;

      if (FieldGroup::is_valid_fid(group->double_fid))
      {
         fids_to_erase.insert(fid);
         fids_to_erase.insert(group->num_fid);
         fids_to_erase.insert(group->denom_fid);
         if (FieldGroup::is_valid_fid(group->float_fid))
            fids_to_erase.insert(group->float_fid);

         if (msg.exists(group->double_fid)) continue;

         whole = num = 0;
         denom = 1;

         msg.getInt(fid, whole);
         msg.getInt(group->num_fid, num);
         msg.getInt(group->denom_fid, denom);

         dbl_fid_vals_to_add.push_back(std::make_pair(group->double_fid, (denom > 0 ? whole + static_cast<double>(num)/denom : 0)));
      }
      else if (FieldGroup::is_valid_fid(group->float_fid))
      {
         fids_to_erase.insert(fid);
         fids_to_erase.insert(group->num_fid);
         fids_to_erase.insert(group->denom_fid);

         whole = num = 0;
         denom = 1;

         msg.getInt(fid, whole);
         msg.getInt(group->num_fid, num);
         msg.getInt(group->denom_fid, denom);

         float_fid_vals_to_add.push_back(std::make_pair(group->float_fid, (denom > 0 ? whole + static_cast<float>(num)/denom: 0)));
      }
   }

   msg.erase(fids_to_erase);

   for (std::vector<std::pair<FieldID, double> >::const_iterator it = dbl_fid_vals_to_add.begin(), it_e = dbl_fid_vals_to_add.end(); it != it_e; ++it)
      msg.addDouble(it->first, it->second);

   for (std::vector<std::pair<FieldID, float> >::const_iterator it = float_fid_vals_to_add.begin(), it_e = float_fid_vals_to_add.end(); it != it_e; ++it)
      msg.addFloat(it->first, it->second);
}

template<class MessageType>
void MD::MD_FloatNormalizer::denormalize_message(MessageType &msg) const
{
#ifdef MD_LIB_64BIT
   std::vector<std::pair<FieldID, int> > wnd_fid_vals_to_add;
   std::set<FieldID> fids_to_erase;
   std::vector<FieldID> all_fids;

   msg.get_field_ids(all_fids);

   double val;
   int32_t whole, num, denom;

   std::vector<FieldID>::const_iterator it = all_fids.begin(), it_e = all_fids.end();
   for (; it != it_e; ++it)
   {
      const FieldID &fid(*it);

      const FieldGroup *group = get_field_group_if_double(fid);
      if (!group) continue;

      if (!FieldGroup::is_valid_fid(group->whole_fid) ||
          !FieldGroup::is_valid_fid(group->num_fid) ||
          !FieldGroup::is_valid_fid(group->denom_fid))
      {
         continue;
      }

      val = 0;
      whole = num = denom = 0;

      msg.getDouble(fid, val);

      if (frac_double(val, whole, num, denom) >= 0)
      {
         wnd_fid_vals_to_add.push_back(std::make_pair(group->whole_fid, whole));
         wnd_fid_vals_to_add.push_back(std::make_pair(group->num_fid, num));
         wnd_fid_vals_to_add.push_back(std::make_pair(group->denom_fid, denom));

         fids_to_erase.insert(fid);

         if (FieldGroup::is_valid_fid(group->float_fid))
            fids_to_erase.insert(group->float_fid);
      }
   }

   // erase fids
   msg.erase(fids_to_erase);

   // add wnd fids
   std::vector<std::pair<FieldID, int> >::const_iterator cit = wnd_fid_vals_to_add.begin(), cit_e = wnd_fid_vals_to_add.end();
   for (; cit != cit_e; ++cit)
      msg.updateInt(cit->first, cit->second);
#endif // MD_LIB_64BIT
}

#endif // __MD_LIB_FLOATNORMALIZER_H__
