/*!******************************************************************
                                                                    
      \filename: ReusableContainer.h
                                                                   
                                                                   
      $DateTime$                                                   
      $Author$                                                     
      $Change$                                                     
      $Id$                                                         
                                                                   
      FactSet Research Systems                                     
      300 First Stamford Place                                     
      Stamford, CT  06902                                          
                                                                  
******************************************************************/
#ifndef __MD_LIB_REUSABLECONTAINER_H__
#define __MD_LIB_REUSABLECONTAINER_H__

#include <algorithm>
#include <atomic>
#include <iterator>
#include <vector>
#include <mutex>

#include <MD/private/prefix.h>

//ReusableContainer is essentially a pool of objects. The objects need to have a default constructor.
//They are taken from the pool with the get function and returned to the pool with the giveBack function.

namespace MD {

template <class T>
class ReusableContainer {
  private:
    std::mutex      m_lock;
    std::vector<T*> m_vFreeList;
  public:
    inline ReusableContainer(size_t uiCapacity = 1)
    {
        std::lock_guard<std::mutex> guard(m_lock);
        m_vFreeList.reserve(uiCapacity);
    }
    ~ReusableContainer()
    {
        std::lock_guard<std::mutex> guard(m_lock);
        while(!m_vFreeList.empty()) {
            delete m_vFreeList.back();
            m_vFreeList.pop_back();
        }
    }
    inline T* get(bool& bFirstUse)
    {
        std::lock_guard<std::mutex> guard(m_lock);
        if(m_vFreeList.empty()) {
            bFirstUse = true;
            return new T;
        } else {
            T* pT = m_vFreeList.back();
            m_vFreeList.pop_back();
            bFirstUse = false;
            return pT;
        }
    }
    inline T* get()
    {
        std::lock_guard<std::mutex> guard(m_lock);
        if(m_vFreeList.empty()) {
            return new T;
        } else {
            T* pT = m_vFreeList.back();
            m_vFreeList.pop_back();
            return pT;
        }
    }
    inline void giveBack(T* pT)
    {
        std::lock_guard<std::mutex> guard(m_lock);
        m_vFreeList.push_back(pT);
    }
};

template < class T>
class ReusableDeleter
{
 private:
  ReusableContainer<T> & m_rContainer;
 public:
  ReusableDeleter(ReusableContainer<T> & rContainer)
    : m_rContainer(rContainer) { }
  void operator() (T * t)
    {
      m_rContainer.giveBack(t);
    }
};

//ReusableContainerWrapper is used for automatically managing the returning of objects to a ReusableContainer.
//It is similar to a scoped_ptr. instead of reset(T*) and reset(), there are set() and unset()

template <class T>
class MD_LIB_API ReusableContainerWrapper
{
private:
  T*                        m_pT;
  std::shared_ptr<ReusableContainer<T>> m_container;
  ReusableContainerWrapper(const ReusableContainerWrapper&);
  ReusableContainerWrapper& operator=(const ReusableContainerWrapper&);

  inline ReusableContainer<T>& rcInstance()
  {
      if(m_container) {
          return *m_container;
      }
      static ReusableContainer<T> m_oRC;
      return m_oRC;
    }

  public:
    inline ReusableContainerWrapper(
        bool                                  bInitialize = false,
        std::shared_ptr<ReusableContainer<T>> container   = nullptr)
        : m_pT(NULL), m_container(container)
    {
        if(bInitialize) {
            set();
        }
    }
    inline ~ReusableContainerWrapper()
    {
        unset();
    }
    inline void set()
    {
        if(!m_pT) {
            m_pT = rcInstance().get();
        }
    }
    inline void unset()
    {
        if(m_pT) {
            rcInstance().giveBack(m_pT);
            m_pT = NULL;
        }
    }
    inline T* get() const
    {
        return m_pT;
    }
    inline T& operator*() const
    {
        return *m_pT;
    }
    inline T * operator->() const
    {
        return m_pT;
    }
    inline void swap(ReusableContainerWrapper & b)
    {
        m_container.swap(b.m_container);
        std::swap(m_pT, b.m_pT);
    }
    inline bool operator!() const
    {
        return (NULL == m_pT);
    }
};

}

#endif // __MD_LIB_REUSABLECONTAINER_H__
