#ifndef __BROADCAST_CLI_H__
#define __BROADCAST_CLI_H__

#include <MD/MD_FieldMap.h>
#include <signal.h>
#include <atomic>
#include <boost/asio.hpp>
#include <condition_variable>
#include <mutex>
#include <fstream>
#include <thread>
#include "Event.h"
#include "FEConsumer.h"
#include "FEConsumerOptions.h"

using namespace FactSet::Datafeed;

class BroadcastCLI {
  public:
    BroadcastCLI();
    ~BroadcastCLI();
    void start(bool console_logging,
               bool print_gaps,
               bool print_messages,
               bool iterate_messages,
               std::vector<std::string>& inputs,
               const std::vector<std::string>& dispatch_prefixes,
               unsigned long slow_threshold,
               bool unsubscribe_on_interrupted);
    bool sigint_handler();
    void on_recovery_message(const std::string& topic,
                             FEConsumer::MsgType& msg);
    void on_message(const std::string& topic, FEConsumer::MsgType& msg);

  private:
    bool m_continue_running;
    std::atomic<bool> m_continue_dispatching;
    bool m_console_logging;
    bool m_print_gaps;
    bool m_print_messages;
    bool m_iterate_messages;
    bool m_unsubscribe_on_interrupted;
    std::vector<std::string> m_dispatch_prefixes;
    std::unique_ptr<FEConsumer> m_consumer;
    std::atomic<uint64_t> m_current_latency;
    std::atomic<uint64_t> m_highest_latency;
    std::atomic<unsigned long> m_msg_count;
    std::atomic<unsigned long> m_interval_msgs;
    std::atomic<unsigned long> m_last_msg_count;
    std::atomic<unsigned long> m_interval_gaps;
    std::atomic<unsigned long> m_gaps_count;
    std::atomic<unsigned long> m_last_gaps_count;
    std::atomic<unsigned long> m_interval_bytes;
    std::atomic<unsigned long> m_bytes_count;
    std::atomic<unsigned long> m_last_bytes_count;
    std::atomic<unsigned long> m_updates;
    boost::asio::io_service m_io_service;
    boost::asio::deadline_timer m_stats_timer;
    bool m_print_stats;
    std::mutex m_lock;
    std::mutex m_message_lock;
    std::unique_ptr<FEConsumer::MsgType> m_slow_message;
    std::string m_slow_topic;
    unsigned long m_slow_time;
    unsigned long m_slow_threshold;
    std::condition_variable m_stats_cv;
    static thread_local std::map<std::string, int> m_sequence_map;
    const MD::MD_FieldMap& m_field_map;
    std::thread m_stats_thread;
    std::vector<std::thread> m_dispatch_threads;
    std::ofstream m_stats_file;
    FEConsumerOptions m_consumer_options;

    void start_stats_thread();

    void quit();
    void create(const FEConsumerOptions& options);
    void create(int sub_threads, int worker_threads);
    void destroy();
    void authenticate(const std::string& host,
                      const std::string& user,
                      const std::string& serial,
                      const std::string& device_id,
                      const std::string& path);
    void connect();
    void disconnect();
    void login(bool update);
    void subscribe(const std::string& topic,
                   bool snapshots_required,
                   FEConsumer::DataMode data_mode);
    void is_subscribed(const std::string& topic);
    void unsubscribe(const std::string& topic);
    void request_file(const std::string& filename);
    void recover(const std::string& topic,
                 time_t begin_time,
                 time_t end_time,
                 FEConsumer::DataMode data_mode);
    void request_snapshot(const std::string& topic,
                 FEConsumer::DataMode data_mode);
    void set_custom_logging(bool enable);
    void set_latency_tracking(bool enable, unsigned long bucket_s);
    void set_heartbeat_timeout(int seconds);
    void set_heartbeat_interval(int seconds);

    void print_help();
    void print_suggest_help();
    void print_consumer_null();
    void print_too_few_args();
    void print_stats();
    void update_stats(const boost::system::error_code& e);
    void iterate_message(const FEConsumer::MsgType& msg);
    void print_message(const std::string& topic,
                       long time_received,
                       const FEConsumer::MsgType& msg);
    bool should_unsubscribe(const Event& e);
    void handle_event(const Event& e);

    static void on_event(const Event& e, void* closure);
    static void on_logging_message(const LogMessage* messages, size_t count);
};
#endif
