#include <boost/program_options.hpp>
#include <boost/program_options/errors.hpp>
#include <string>
#include <iostream>
#include "FEConsumer.h"
#include "FEConsumerOptions.h"
#include "BroadcastFields.h"


using namespace FactSet::Datafeed;

//Example for running Replay functionality

void on_message(const std::string &topic, FEConsumer::MsgType &msg, void *closure) {
 // Get sequence number
 int seq = -1;
 if(!msg.getInt(BroadcastFieldId::FDS_SEQ_NUM, seq)) {
   std::cout << "Message for " << topic << " missing sequence number"
   << std::endl;
   return;
 }
 std::cout << "Message for " << topic << " with sequence number " << seq
 << std::endl;
 std::cout << msg << std::endl;
}

int main(int argc, char* argv[]) {

   const std::string default_connect_host = "tcp://localhost:5500";

   std::string connect_host;
   std::vector<std::string> topics;

   namespace po = boost::program_options;
   po::options_description description("Options");
   description.add_options()(
      "connect_host,c", po::value<std::string>(&connect_host)->default_value(default_connect_host),
      std::string("Host to connect to High Throughput Test Service. e.g. ").append(default_connect_host).c_str());
   description.add_options()(
      "topics,t", po::value<std::vector<std::string>>(&topics)
      ->required()
      ->composing()
      ->multitoken(),
      "Topics to subscribe to in format <product|issue|iso|symbol|> e.g. 9001| or 9001|1|USA|FDS|");
    description.add_options()("help,h", "Display help");

    po::variables_map vm;

    try {
      po::store(po::parse_command_line(argc, argv, description), vm);
      po::notify(vm);
    } catch (po::error_with_option_name &e) {
      if (vm.count("help")) {
        std::cout << description << std::endl;
        std::cout << "Example usage: " << "./simple -c " << default_connect_host << " -t 9005 9001 9002" << std::endl; 
        return 0;
      } else {
        std::cout << e.what() << std::endl;
        return 0;
      }
    }


    std::cout << "--------------------------------------------" << std::endl;
    std::cout << "FactSet Research Systems " << std::endl;
    std::cout << "Starting process [" << argv[0] << "]" << std::endl;
    std::cout << "Connect Host            : " << connect_host << std::endl;
    std::cout << "Topics                  : ";
    for(auto& topic: topics){
       std::cout << topic << " ";
    }
    std::cout << std::endl;


   // Create consumer with 1 subscriber and 1 worker
   FEConsumerOptions options;
   options.set_mode(FEConsumerOptions::Mode::REPLAY, connect_host);

   FEConsumer consumer(options);
      
   consumer.enable_verbose_logging();
   Error err = consumer.connect();
   if(err != Error::NoError) {
      std::cout << std::endl << "Error connecting on " << connect_host << std::endl;
      return -1;
   }

   // Set a message callback for any topic
   FEConsumer::SocketType fd;
   err = consumer.register_callback("", &on_message, nullptr, fd);
   if(err != Error::NoError) {
      std::cout << std::endl << "Error setting callback for all topics " << connect_host << std::endl;
      return -1;
   }

   // Subscribe to topics
   for(auto& topic: topics){
      err = consumer.subscribe(topic, FEConsumer::NO_SNAPSHOT, FEConsumer::LIVE);
      if(err != Error::NoError) {
         std::cout << std::endl << "Skipping: Error subscribing to " << topic << std::endl;
      }
   }

   // Dispatch with 1000ms timeout
   while(true) {
      consumer.dispatch(1000);
   }

   return 0;
}